#!/home/gchossie/anaconda3/envs/freight_us/bin/python

#SBATCH --cpus-per-task=24
#SBATCH --mem=0
#SBATCH --time=05:00:00

import geopandas as gpd
import rioxarray
from shapely.geometry import box, mapping
import pickle
import xarray as xr
import numpy as np
from sklearn.neighbors import BallTree
import sys
import pandas as pd
import pykrige.kriging_tools as kt
from pykrige.ok import OrdinaryKriging
from joblib import Parallel, delayed
import multiprocessing

if __name__ == '__main__':
    ensemble = sys.argv[1]
else:
    ensemble = 'Europe'

PATH_LEGACY = '/home/gchossie/2020_aq/monitor_based/'
TARGET_DIR = '/net/fs03/d1/gchossie/2020_aq/monitor_data/'
if TARGET_DIR == '':
    datadir = './'
else:
    datadir = TARGET_DIR

all_regs = gpd.read_file(PATH_LEGACY + 'data/all_regions.geojson')

def get_path_raw_vals(ensemble):
    if ensemble == 'China':
        return '/home/gchossie/2020_aq/monitor_based/data/ch_raw/'
    if ensemble == 'Korea':
        return '/home/gchossie/2020_aq/monitor_based/data/kr_raw/'
    if ensemble == 'Japan':
        return '/home/gchossie/2020_aq/monitor_based/data/jp_raw/'
    if ensemble == 'United States':
        return '/home/gchossie/2020_aq/monitor_based/data/us_raw/'
    if ensemble == 'Europe':
        return '/net/fs03/d1/gchossie/2020_aq/data/eu_raw/'

def get_pop(ensemble):
    pop15 = None
    if ensemble == 'China':
        with open('./data/ch_raw/pop_ch.pkl', 'rb') as f:
            pop15 = pickle.load(f)
        pop15 = pop15.rio.set_crs('epsg:4326')
        pop15 = pop15.rio.set_spatial_dims('lon', 'lat')
    elif ensemble == 'Korea':
        with open('./data/kr_raw/pop_kr_full_res.pkl', 'rb') as f:
            pop15 = pickle.load(f)
        pop15 = pop15.rio.set_crs('epsg:4326')
        pop15 = pop15.rio.set_spatial_dims('lon', 'lat')
    elif ensemble == 'Japan':
        with open('./data/jp_raw/pop_jp_full_res.pkl', 'rb') as f:
            pop15 = pickle.load(f)
        pop15 = pop15.rio.set_crs('epsg:4326')
        pop15 = pop15.rio.set_spatial_dims('lon', 'lat')
    elif ensemble == 'United States':
        with open('./data/us_raw/pop_us_0.125.pkl', 'rb') as f:
            pop15 = pickle.load(f)
        pop15 = pop15.rio.set_crs('epsg:4326')
        pop15 = pop15.rio.set_spatial_dims('lon', 'lat')
    elif ensemble == 'Europe':
        with open('/net/fs03/d1/gchossie/2020_aq/data/eu_raw/pop_eu.pkl', 'rb') as f:
            pop15 = pickle.load(f)
        pop15 = pop15.rio.set_crs('epsg:4326')
        pop15 = pop15.rio.set_spatial_dims('lon', 'lat')
    return pop15

def get_grid(ensemble):
    pop15 = get_pop(ensemble)
    regions = all_regs.loc[all_regs.Ensemble == ensemble]
    regs = list(set(regions['Region']))
    try:
        regs.remove('Ceuta y Melilla')
    except:
        pass
    try:
        regs.remove('Islas Canarias')
    except:
        pass
    mask = xr.DataArray(np.nan,
                        coords=[pop15['lat'].values, pop15['lon'].values],
                        dims=['lat', 'lon'])
    popmaps = dict.fromkeys(regs)
    for k in popmaps.keys():
        popmaps[k] = xr.DataArray(np.nan,
                                  coords=[pop15['lat'].values,
                                          pop15['lon'].values],
                                  dims=['lat', 'lon'])
    # For each state, clip the population, then put into the big mask dataarray
    for stname in regs:
        reg = regions.loc[regions['Region'] == stname]
        reg.crs = 'epsg:4326'
        try:
            tmp = pop15.rio.clip(reg.geometry.apply(mapping), reg.crs)
        except:
            continue
        popmaps[stname] = popmaps[stname].combine_first(tmp)
        tmp.values[~np.isnan(tmp.values)] = np.where(
            tmp.values[~np.isnan(tmp.values)] > 0, 1, 0)
        mask = mask.combine_first(tmp)
    mask.values = 1 - np.nan_to_num(mask.values)
    mask.values = mask.values.astype(bool)

    return mask, regs, popmaps

class tree(object):
    def __init__(self, X=None, z=None, leafsize=10):
        if not X is None:
            self.tree = BallTree(X, leaf_size=leafsize, metric='haversine')
        if not z is None:
            self.z = np.array(z)

    def fit(self, X=None, z=None, leafsize=10):
        return self.__init__(X, z, leafsize)

    def __call__(self, X, k=20, eps=1e-6, p=2, regularize_by=1e-9):
        self.distances, self.idx = self.tree.query(X, k, return_distance=True)
        self.distances += regularize_by
        weights = self.z[self.idx.ravel()].reshape(self.idx.shape)
        mw = np.sum(weights/(self.distances**2), axis=1) / np.sum(1./(self.distances**2), axis=1)
        return mw

    def transform(self, X, k=20, p=2, eps=1e-6, regularize_by=1e-9):
        return self.__call__(X, k, eps, p, regularize_by)

def get_kriging_params(ensemble,sp):
    if ensemble == 'Europe':
        if sp == 'NO2': return {'sill': 2.0, 'range': 7.5, 'nugget': 0.1}
        elif sp == 'O3': return {'sill': 3.0, 'range': 2.5, 'nugget': 0.1}
        else: return {'sill': 4.0, 'range': 6.5, 'nugget': 0.1}
    elif ensemble == 'United States':
        if sp == 'NO2': return {'sill': 3.0, 'range': 6.5, 'nugget': 0.1}
        elif sp == 'O3': return {'sill': 2.0, 'range': 7.5, 'nugget': 0.1}
        else: return {'sill': 2.0, 'range': 4.5, 'nugget': 0.1}
    elif ensemble == 'Korea':
        if sp == 'NO2': return {'sill': 2.0, 'range': 9.5, 'nugget': 0.1}
        elif sp == 'O3': return {'sill': 2.0, 'range': 2.5, 'nugget': 0.1}
        else: return {'sill': 2.0, 'range': 2.5, 'nugget': 0.1}
    elif ensemble == 'China':
        if sp == 'NO2': return {'sill': 2.0, 'range': 4.5, 'nugget': 0.1}
        elif sp == 'O3': return {'sill': 5.0, 'range': 5.5, 'nugget': 0.1}
        else: return {'sill': 3.0, 'range': 8.5, 'nugget': 0.1}
    elif ensemble == 'Japan':
        if sp == 'NO2': return {'sill': 2.0, 'range': 9.5, 'nugget': 0.1}
        elif sp == 'O3': return {'sill': 2.0, 'range': 2.5, 'nugget': 0.1}
        else: return {'sill': 2.0, 'range': 2.5, 'nugget': 0.1}
    else: return None

def preprocess_raw_vals(sp):
    '''

    '''
    PATH = '/net/fs03/d1/gchossie/2020_aq/monitor_data/raw_data/'
    vals = pd.read_pickle(PATH + f'{sp}_all_measurements.pkl')
    vals.dropna(axis=0, how='all', inplace=True)
    vals = vals.reset_index().rename(columns={'index': 'Identifier'})
    monitors = pd.read_feather(PATH + f'{sp}_monitor_locations.feather')
    vals = pd.merge(vals, monitors[['Identifier', 'Latitude', 'Longitude']], on='Identifier').set_index('Identifier')
    return vals
    
def get_interpolated_dat(ensemble, dat, method='kriging', rawvals=None):
    '''

    rawvals = dict indexed by species of preprocessed df with all measurements
    '''
        # Shortcut the large grid compuation
    with open(PATH_LEGACY + 'data/masks_ensembles.pkl','rb') as f:
        masks = pickle.load(f)
    with open(PATH_LEGACY + 'data/regs_ensembles.pkl','rb') as f:
        regnames = pickle.load(f)
    with open(PATH_LEGACY + 'data/popmaps_ensembles.pkl','rb') as f:
        popumaps = pickle.load(f)
    lst_issues = {'no data for this sp': [], 'no data at this dat': [],
                  'failed to get exposure': []}
    if not os.path.isdir(datadir + 'ts_interpolated'):
        os.mkdir(datadir + 'ts_interpolated')
    datt = dat.strftime(format='%Y%m%d')
    species = ['O3', 'PM2.5', 'NO2']
    #out 

    mask, regs, popmaps = masks[ensemble], regnames[ensemble], popumaps[ensemble]
    gridlon = mask['lon'].values
    gridlat = mask['lat'].values
    if rawvals is None:
        rawvals = dict.fromkeys(species)
        for sp in species:
            rawvals[sp] = preprocess_raw_vals(sp)
    for sp in species:
        
        try:
            valstemp = rawvals[~pd.isnull(rawvals['Latitude'])]
        except:
            print(f'{sp} at {dat} not found')
            lst_issues['no data at this dat'].append((dat,sp))
            continue
        if dat not in valstemp.columns:
            print(f'{sp} at {dat} not found')
            lst_issues['no data at this dat'].append((dat,sp))
            continue

        valstemp = valstemp[~pd.isnull(valstemp['Latitude'])]
        valstemp = valstemp.loc[~pd.isnull(valstemp[dat])]
        valstemp = valstemp.loc[valstemp[dat] > 0].reset_index(drop=True)

        if len(valstemp) == 0:
            print(f'{sp} on {dat} has no values')
            lst_issues['no data for this sp'].append((dat, sp))
            continue

        if method == 'kriging':
            params = get_kriging_params(ensemble,sp)
            data = valstemp[['Longitude', 'Latitude', dat]].values
    
            OK = OrdinaryKriging(data[:, 0], data[:, 1], data[:, 2],
                                 variogram_model='spherical',
                                 variogram_parameters=params,
                                 verbose=False, enable_plotting=False,
                                 coordinates_type='geographic')
    
            try:
                z2, ss = OK.execute('masked', mask['lon'].values,
                                   mask['lat'].values, mask=mask.values, backend='C',
                                   n_closest_points=80)
            except:
                z2, ss = OK.execute(
                    'masked', mask['lon'].values, mask['lat'].values,
                    mask=mask.values, backend='C')

            z2 = np.ma.filled(z2, fill_value=0.0)

        elif method == 'idw': 
            data = valstemp[['Latitude','Longitude',dat]].values
            # Train 
            idw_tree = tree(data[:, :2] * np.pi/180, data[:, 2])
            # Predict
            gridlon_rad, gridlat_rad = gridlon * np.pi/180, gridlat * np.pi/180
            X2 = np.meshgrid(gridlat_rad,gridlon_rad)
            grid_shape = X2[0].shape
            X2 = np.reshape(X2, (2,-1)).T
            try:
                z2 = idw_tree(X2)
            except:
                z2 = idw_tree(X2,k=2)
            z2 = z2.reshape(grid_shape).T

        krgrid = xr.DataArray(
            data=z2, coords=[gridlat,gridlon],
            dims=['lat', 'lon'])
        for regname in regs:
            try:
                out = pd.Series(index=['mean', 'low', 'high'], dtype=float)
                out['mean'] = float((krgrid * popmaps[regname]).sum().values) / float(popmaps[regname].sum())
            except:
                print(f'Problem with {sp} in {regname} at {dat}.')
                lst_issues['failed to get exposure'].append(
                    (dat, sp, regname))
                continue
            with open(datadir + f'ts_interpolated/{sp}_{regname}_{datt}_{method}.pkl', 'wb') as f:
                pickle.dump(out, f)
    with open(datadir + f'ts_interpolated/issues_{ensemble}_{datt}.pkl', 'wb') as f:
        pickle.dump(lst_issues,f)
    return float(out['mean'])

if __name__ == '__main__':
    num_cores = multiprocessing.cpu_count()
    results = Parallel(n_jobs=num_cores)(delayed(get_interpolated_dat)(ensemble,dat,method='kriging')
                                     for dat in pd.date_range('20160101', '20200625', freq='D'))



